@@ -2,6 +2,14 @@
 // Microsoft Developer Studio generated include file.
 // Used by resource.rc
 //
+#define IDS_FAIL_CREATEFILE             1
+#define IDS_FAIL_MAPVIEWOFFILE          3
+#define IDS_SCRIPT_NOT_FOUND            4
+#define IDS_FAIL_MODIFY_IMPORT_TABLE    5
+#define IDS_FAIL_GETPROCADDRESS         6
+#define IDS_CANNOT_EXECUTE_SCRIPT       7
+#define IDS_CANNOT_OBTAIN_EXCEPTION     8
+#define IDS_HEADER_NOT_FOUND            9
 #define IDD_EXCEPTION                   101
 #define IDI_RUBY                        102
 #define IDI_EXCEPTION                   103
