@@ -1,4 +1,4 @@
-// $Id: exerb.cpp,v 1.56 2002/11/15 07:16:07 yuya Exp $
+// $Id: exerb.cpp,v 1.63 2002/11/27 07:17:50 yuya Exp $
 
 ////////////////////////////////////////////////////////////////////////////////
 
@@ -35,9 +35,9 @@ using namespace std;
 #endif
 
 #ifdef _DEBUG
-#define DEBUGMSG0(msg)          ::fprintf(stderr, msg)
-#define DEBUGMSG1(msg, p1)      ::fprintf(stderr, msg, p1)
-#define DEBUGMSG2(msg, p1, p2)  ::fprintf(stderr, msg, p1, p2)
+#define DEBUGMSG0(msg)          ::fprintf(stderr, "[Exerb] " msg)
+#define DEBUGMSG1(msg, p1)      ::fprintf(stderr, "[Exerb] " msg, p1)
+#define DEBUGMSG2(msg, p1, p2)  ::fprintf(stderr, "[Exerb] " msg, p1, p2)
 #else
 #define DEBUGMSG0(msg)
 #define DEBUGMSG1(msg, p1)
@@ -78,8 +78,6 @@ static LRESULT CALLBACK ExFailDialogProc(HWND hWnd, UINT uiMessage, WPARAM wPara
 static void  ExErrorMessage(int iResourceID);
 static void  ExErrorMessage(LPSTR pszMessage);
 static void  ExRaiseLoadError(int iResourceID);
-static VALUE ExRequireWithExtension(VALUE vFileName, char *pszExtension);
-static VALUE ExRequireWithoutExtension(VALUE vFileName);
 static void  ExCreateTemporaryFile(char *pszFilePathBuffer);
 static bool  ExGetImportTable(DWORD dwDosHeader, DWORD *pdwAddress, DWORD *pdwDelta);
 static PIMAGE_SECTION_HEADER ExGetEnclosingSectionHeader(PIMAGE_NT_HEADERS pNtHeader, DWORD dwRVA);
@@ -88,7 +86,6 @@ static PIMAGE_IMPORT_DESCRIPTOR ExGetFirstImportDescriptor(DWORD dwDosHeader, DW
 static void  ExReplaceImportDllName(DWORD dwDosHeader, DWORD dwImportTableDelta, PDWORD pdwNamePoolAddress, PDWORD pdwNamePoolSize, PIMAGE_IMPORT_DESCRIPTOR pFirstDescriptor, char* pszSrc, char* pszDest);
 static void  ExReplaceImportFunctionName(DWORD dwOffsetOfName, PIMAGE_IMPORT_DESCRIPTOR pFirstDescriptor, const char* pszDllName, const char* pszSource, const char* pszDestination);
 static bool  ExGetSectionUnusedArea(PIMAGE_NT_HEADERS pNtHeader, char* pszName, PDWORD pdwAddress, PDWORD pdwSize);
-//static void  ExGetNamePool(PIMAGE_NT_HEADERS pNtHeader, PDWORD pdwAddress, PDWORD pdwSize);
 static void  ExReplaceImportTable(DWORD dwOffset);
 static void  ExDllWriteToFile(DWORD dwOffset, DWORD dwSize, char *pszFilePath);
 static HMODULE ExLoadDll(char *pszFilePath, VALUE vFileName);
@@ -98,6 +95,7 @@ static VALUE ExRequireScript(VALUE vFileName, DWORD dwID);
 static void  ExGetSelfFilePath(LPSTR pszBuffer, UINT uiBufferSize);
 static void  ExGetSelfFileName(LPSTR pszBuffer, UINT uiBufferSize);
 static VALUE ExEvalString(VALUE vString, VALUE vFileName);
+static bool  ExFindFile(VALUE vFilename, bool *pbShared, VALUE *pvRealname, DWORD *pdwID);
 static VALUE ExRequire(VALUE valueObject, VALUE valueFileName);
 
 static VALUE rb_exerb_s_runtime_p(VALUE self);
@@ -121,6 +119,8 @@ typedef struct {
 
 ////////////////////////////////////////////////////////////////////////////////
 
+static VALUE g_vOldRequire;
+
 static void* g_pArchive = NULL;
 
 static DWORD g_dwBaseOfNameTable   = 0;
@@ -196,7 +196,7 @@ ExInitModule()
 static void
 ExReplaceRequire(void)
 {
-	::rb_alias(rb_mKernel, ::rb_intern("__old_require__"), ::rb_intern("require"));
+	g_vOldRequire = ::rb_funcall(rb_mKernel, ::rb_intern("method"), 1, ID2SYM(::rb_intern("require")));
 	::rb_define_global_function("require", (RUBYFUNC)ExRequire, 1);
 }
 
@@ -209,7 +209,7 @@ ExMapping(void)
 	    return false;
 	}
 
-	int nArchiveSize = (pDosHeader->SizeOfCompressedArchive > 0 ? pDosHeader->SizeOfCompressedArchive : pDosHeader->SizeOfArchive);
+	const int nArchiveSize = (pDosHeader->SizeOfCompressedArchive > 0 ? pDosHeader->SizeOfCompressedArchive : pDosHeader->SizeOfArchive);
 	g_pArchive = new char[nArchiveSize];
 
 	char szFilePath[MAX_PATH] = "";
@@ -334,9 +334,9 @@ ExExecuteFirstScript(VALUE data)
 ExExecuteFirstScript(void)
 #endif
 {
-	PEXERB_SCRIPT_HEADER pScriptHeader = (PEXERB_SCRIPT_HEADER)g_dwBaseOfScriptTable;
-	VALUE vScirptSource = ::rb_str_new((char*)(g_dwBaseOfScriptTable + pScriptHeader->Offset), pScriptHeader->Size);
-	VALUE vScritpName   = ::rb_str_new2(g_Id2NameTable[pScriptHeader->NameID].c_str());
+	const PEXERB_SCRIPT_HEADER pScriptHeader = (PEXERB_SCRIPT_HEADER)g_dwBaseOfScriptTable;
+	const VALUE vScirptSource = ::rb_str_new((char*)(g_dwBaseOfScriptTable + pScriptHeader->Offset), pScriptHeader->Size);
+	const VALUE vScritpName   = ::rb_str_new2(g_Id2NameTable[pScriptHeader->NameID].c_str());
 	::ruby_script(STR2CSTR(vScritpName));
 	::ExEvalString(vScirptSource, vScritpName);
 
@@ -394,11 +394,11 @@ ExFail(void)
 static void
 ExPrintStackTrace()
 {
-	VALUE type       = ::rb_funcall(::rb_funcall(ruby_errinfo, ::rb_intern("type"), 0), ::rb_intern("name"), 0);
-	VALUE message    = ::rb_funcall(ruby_errinfo, ::rb_intern("message"), 0);
-	VALUE backtrace  = ::rb_funcall(ruby_errinfo, ::rb_intern("backtrace"), 0);
-	VALUE backtrace1 = ::rb_ary_shift(backtrace);
-	VALUE backtrace2 = ::rb_str_concat(::rb_str_new2("\tfrom "), ::rb_ary_join(backtrace, ::rb_str_new2("\n\tfrom ")));
+	const VALUE type       = ::rb_funcall(::rb_funcall(ruby_errinfo, ::rb_intern("type"), 0), ::rb_intern("name"), 0);
+	const VALUE message    = ::rb_funcall(ruby_errinfo, ::rb_intern("message"), 0);
+	const VALUE backtrace  = ::rb_funcall(ruby_errinfo, ::rb_intern("backtrace"), 0);
+	const VALUE backtrace1 = ::rb_ary_shift(backtrace);
+	const VALUE backtrace2 = ::rb_str_concat(::rb_str_new2("\tfrom "), ::rb_ary_join(backtrace, ::rb_str_new2("\n\tfrom ")));
 
 	::fprintf(stderr, "%s: %s (%s)\n", STR2CSTR(backtrace1), STR2CSTR(message), STR2CSTR(type));
 	if ( FIX2INT(::rb_funcall(backtrace, ::rb_intern("size"), 0)) > 0 ) {
@@ -412,12 +412,12 @@ ExFailDialogProc(HWND hWnd, UINT uiMessage, WPARAM wParam, LPARAM lParam)
 	switch ( uiMessage ) {
 		case WM_INITDIALOG:
 			if ( ruby_errinfo ) {
-				VALUE vType            = ::rb_funcall(ruby_errinfo, ::rb_intern("type"), 0);
-				VALUE vTypeName        = ::rb_funcall(vType, ::rb_intern("name"), 0);
-				VALUE vMessage         = ::rb_funcall(ruby_errinfo, ::rb_intern("message"), 0);
-				VALUE vMessageString   = ::rb_funcall(vMessage, ::rb_intern("gsub"), 2, ::rb_str_new2("\n"), ::rb_str_new2("\r\n"));
-				VALUE vBacktrace       = ::rb_funcall(ruby_errinfo, ::rb_intern("backtrace"), 0);
-				VALUE vBacktraceString = ::rb_str_concat(::rb_ary_join(vBacktrace, ::rb_str_new2("\r\n")), rb_str_new2("\r\n"));
+				const VALUE vType            = ::rb_funcall(ruby_errinfo, ::rb_intern("type"), 0);
+				const VALUE vTypeName        = ::rb_funcall(vType, ::rb_intern("name"), 0);
+				const VALUE vMessage         = ::rb_funcall(ruby_errinfo, ::rb_intern("message"), 0);
+				const VALUE vMessageString   = ::rb_funcall(vMessage, ::rb_intern("gsub"), 2, ::rb_str_new2("\n"), ::rb_str_new2("\r\n"));
+				const VALUE vBacktrace       = ::rb_funcall(ruby_errinfo, ::rb_intern("backtrace"), 0);
+				const VALUE vBacktraceString = ::rb_str_concat(::rb_ary_join(vBacktrace, ::rb_str_new2("\r\n")), rb_str_new2("\r\n"));
 
 				::SetDlgItemText(hWnd, IDC_EDIT_TYPE,      STR2CSTR(vTypeName));
 				::SetDlgItemText(hWnd, IDC_EDIT_MESSAGE,   STR2CSTR(vMessageString));
@@ -435,7 +435,7 @@ ExFailDialogProc(HWND hWnd, UINT uiMessage, WPARAM wParam, LPARAM lParam)
 			}
 
 			{
-				HFONT hFont = ::CreateFont(14, 0, 0, 0, FW_REGULAR, FALSE, FALSE, FALSE, SHIFTJIS_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, PROOF_QUALITY, FIXED_PITCH | FF_MODERN, "Terminal");				SetWindowFont(::GetDlgItem(hWnd, IDC_EDIT_TYPE),      hFont, false);
+				const HFONT hFont = ::CreateFont(14, 0, 0, 0, FW_REGULAR, FALSE, FALSE, FALSE, SHIFTJIS_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, PROOF_QUALITY, FIXED_PITCH | FF_MODERN, "Terminal");				SetWindowFont(::GetDlgItem(hWnd, IDC_EDIT_TYPE),      hFont, false);
 				SetWindowFont(::GetDlgItem(hWnd, IDC_EDIT_MESSAGE),   hFont, false);
 				SetWindowFont(::GetDlgItem(hWnd, IDC_EDIT_BACKTRACE), hFont, false);
 			}
@@ -487,53 +487,6 @@ ExRaiseLoadError(int iResourceID)
 	::rb_raise(rb_eLoadError, message);
 }
 
-static VALUE
-ExRequireWithExtension(VALUE vFileName, char *pszExtension)
-{
-	DWORD dwID = g_Name2IdTable[STR2CSTR(vFileName)];
-
-	if ( dwID ) {
-		if ( ::stricmp(".rb", pszExtension) == 0 ) {
-			return ::ExRequireScript(vFileName, dwID);
-		} else if ( ::stricmp(".so", pszExtension) == 0 || ::stricmp(".dll", pszExtension) == 0 ) {
-			return ::ExRequireDll(vFileName, dwID);
-		} else {
-			return Qfalse;
-		}
-	} else {
-		return Qfalse;
-	}
-}
-
-static VALUE
-ExRequireWithoutExtension(VALUE vFileName)
-{
-	VALUE vFileNameRB  = ::rb_str_concat(::rb_str_dup(vFileName), ::rb_str_new2(".rb"));
-	VALUE vFileNameSO  = ::rb_str_concat(::rb_str_dup(vFileName), ::rb_str_new2(".so"));
-	VALUE vFileNameDLL = ::rb_str_concat(::rb_str_dup(vFileName), ::rb_str_new2(".dll"));
-
-	const DWORD idRB  = g_Name2IdTable[STR2CSTR(vFileNameRB)];
-	const DWORD idSO  = g_Name2IdTable[STR2CSTR(vFileNameSO)];
-	const DWORD idDLL = g_Name2IdTable[STR2CSTR(vFileNameDLL)];
-
-/*
-	const char* filename = STR2CSTR(vFileName);
-	const DWORD idRB  = g_Name2IdTable[string(filename).append(".rb")];
-	const DWORD idSO  = g_Name2IdTable[string(filename).append(".so")];
-	const DWORD idDLL = g_Name2IdTable[string(filename).append(".dll")];
-*/
-
-	if ( idRB ) {
-		return ::ExRequireScript(vFileNameRB, idRB);
-	} else if ( idSO ) {
-		return ::ExRequireDll(vFileNameSO, idSO);
-	} else if ( idDLL ) {
-		return ::ExRequireDll(vFileNameDLL, idDLL);
-	} else {
-		return Qfalse;
-	}
-}
-
 static void
 ExCreateTemporaryFile(char *pszFilePathBuffer)
 {
@@ -545,11 +498,12 @@ ExCreateTemporaryFile(char *pszFilePathBuffer)
 static bool
 ExGetImportTable(DWORD dwDosHeader, DWORD *pdwAddress, DWORD *pdwDelta)
 {
-	PIMAGE_DOS_HEADER pDosHeader = (PIMAGE_DOS_HEADER)dwDosHeader;
-	PIMAGE_NT_HEADERS pNtHeader  = ::ExGetNtHeader(pDosHeader);
-	DWORD dwImportTableRVA = pNtHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress;
+	const PIMAGE_DOS_HEADER pDosHeader = (PIMAGE_DOS_HEADER)dwDosHeader;
+	const PIMAGE_NT_HEADERS pNtHeader  = ::ExGetNtHeader(pDosHeader);
 
-	PIMAGE_SECTION_HEADER pSection = ::ExGetEnclosingSectionHeader(pNtHeader, dwImportTableRVA);
+	const DWORD dwImportTableRVA = pNtHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress;
+
+	const PIMAGE_SECTION_HEADER pSection = ::ExGetEnclosingSectionHeader(pNtHeader, dwImportTableRVA);
 	if ( pSection ) {
 		*pdwDelta   = pSection->VirtualAddress - pSection->PointerToRawData;
 		*pdwAddress = dwImportTableRVA - *pdwDelta;
@@ -593,6 +547,7 @@ static PIMAGE_IMPORT_DESCRIPTOR
 ExGetFirstImportDescriptor(DWORD dwDosHeader, DWORD *pdwImportTableDelta)
 {
 	DWORD dwImportTableAddress = 0;
+
 	if ( ::ExGetImportTable(dwDosHeader, &dwImportTableAddress, pdwImportTableDelta) ) {
 		DWORD dwImportTableBase = dwDosHeader + dwImportTableAddress;
 		return (PIMAGE_IMPORT_DESCRIPTOR)dwImportTableBase;
@@ -604,8 +559,8 @@ ExGetFirstImportDescriptor(DWORD dwDosHeader, DWORD *pdwImportTableDelta)
 static void
 ExReplaceImportDllName(DWORD dwDosHeader, DWORD dwImportTableDelta, PDWORD pdwNamePoolAddress, PDWORD pdwNamePoolSize, PIMAGE_IMPORT_DESCRIPTOR pFirstDescriptor, char* pszSrc, char* pszDest)
 {
-	DWORD dwOffsetOfName = dwDosHeader - dwImportTableDelta;
-	DWORD dwDestLength   = ::strlen(pszDest);
+	const DWORD dwOffsetOfName = dwDosHeader - dwImportTableDelta;
+	const DWORD dwDestLength   = ::strlen(pszDest);
 
 	for ( PIMAGE_IMPORT_DESCRIPTOR pDescriptor = pFirstDescriptor; pDescriptor->Name; pDescriptor++ ) {
 		char *pszName = (char*)(dwOffsetOfName + pDescriptor->Name);
@@ -614,7 +569,7 @@ ExReplaceImportDllName(DWORD dwDosHeader, DWORD dwImportTableDelta, PDWORD pdwNa
 			if ( dwDestLength <= ::strlen(pszName) ) {
 				::strcpy(pszName, pszDest);
 			} else if ( dwDestLength + 1 <= *pdwNamePoolSize ) {
-				DWORD dwAddress = *pdwNamePoolAddress - dwDestLength - 1;
+				const DWORD dwAddress = *pdwNamePoolAddress - dwDestLength - 1;
 				pDescriptor->Name = dwAddress + dwImportTableDelta;
 
 				::memcpy((void*)(dwDosHeader + dwAddress), pszDest, dwDestLength);
@@ -632,7 +587,7 @@ static void
 ExReplaceImportFunctionName(DWORD dwOffsetOfName, PIMAGE_IMPORT_DESCRIPTOR pFirstDescriptor, const char* pszDllName, const char* pszSource, const char* pszDestination)
 {
 	for ( PIMAGE_IMPORT_DESCRIPTOR pDescriptor = pFirstDescriptor; pDescriptor->Name; pDescriptor++ ) {
-		char *pszName = (char*)(dwOffsetOfName + pDescriptor->Name);
+		const char *pszName = (char*)(dwOffsetOfName + pDescriptor->Name);
 
 		if ( ::strcmp(pszName, pszDllName) == 0 ) {
 			PIMAGE_THUNK_DATA thunk    = (PIMAGE_THUNK_DATA)pDescriptor->Characteristics;
@@ -650,8 +605,8 @@ ExReplaceImportFunctionName(DWORD dwOffsetOfName, PIMAGE_IMPORT_DESCRIPTOR pFirs
 
 			while ( thunk->u1.AddressOfData ) {
 				if ( !(thunk->u1.Ordinal & IMAGE_ORDINAL_FLAG) ) {
-					PIMAGE_IMPORT_BY_NAME pImportByName = (PIMAGE_IMPORT_BY_NAME)((DWORD)(thunk->u1.AddressOfData) + dwOffsetOfName);
-					LPSTR pszFunctionName = (LPSTR)pImportByName->Name;
+					const PIMAGE_IMPORT_BY_NAME pImportByName = (PIMAGE_IMPORT_BY_NAME)((DWORD)(thunk->u1.AddressOfData) + dwOffsetOfName);
+					const LPSTR pszFunctionName = (LPSTR)pImportByName->Name;
 					if ( ::strcmp(pszFunctionName, pszSource) == 0 ) {
 						::strcpy(pszFunctionName, pszDestination);
 					}
@@ -667,7 +622,7 @@ ExReplaceImportFunctionName(DWORD dwOffsetOfName, PIMAGE_IMPORT_DESCRIPTOR pFirs
 static bool
 ExGetSectionUnusedArea(PIMAGE_NT_HEADERS pNtHeader, char* pszName, PDWORD pdwAddress, PDWORD pdwSize)
 {
-	PIMAGE_SECTION_HEADER pSection = ::ExFindSection(pNtHeader, pszName);
+	const PIMAGE_SECTION_HEADER pSection = ::ExFindSection(pNtHeader, pszName);
 
 	if ( pSection ) {
 		*pdwAddress = pSection->PointerToRawData + pSection->SizeOfRawData;
@@ -680,43 +635,22 @@ ExGetSectionUnusedArea(PIMAGE_NT_HEADERS pNtHeader, char* pszName, PDWORD pdwAdd
 	}
 }
 
-/*
-static void
-ExGetNamePool(PIMAGE_NT_HEADERS pNtHeader, PDWORD pdwAddress, PDWORD pdwSize)
-{
-	char* ppszSections[] = {".idata", ".edata", ".rsrc"};
-	int   nSections      = 3;
-	DWORD dwAddress = 0;
-	DWORD dwSize    = 0;
-
-	for ( int i = 0; i < nSections; i++ ) {
-		if ( ::ExGetSectionUnusedArea(pNtHeader, ppszSections[i], &dwAddress, &dwSize) ) {
-			if ( *pdwSize < dwSize ) {
-				*pdwAddress = dwAddress;
-				*pdwSize    = dwSize;
-			}
-		}
-	}
-}
-*/
-
 static void
 ExReplaceImportTable(DWORD dwOffset)
 {
-	DWORD dwDosHeader = dwOffset;
+	const DWORD dwDosHeader = dwOffset;
 
 	DWORD dwImportTableDelta = 0;
-	PIMAGE_IMPORT_DESCRIPTOR pDescriptor = ::ExGetFirstImportDescriptor(dwDosHeader, &dwImportTableDelta);
+	const PIMAGE_IMPORT_DESCRIPTOR pDescriptor = ::ExGetFirstImportDescriptor(dwDosHeader, &dwImportTableDelta);
 	if ( !pDescriptor ) {
 		::ExRaiseLoadError(IDS_FAIL_MODIFY_IMPORT_TABLE);
 	}
 
 	DWORD dwNamePoolAddress = 0;
 	DWORD dwNamePoolSize    = 0;
-	PIMAGE_NT_HEADERS     pNtHeader      = ::ExGetNtHeader((PIMAGE_DOS_HEADER)dwDosHeader);
+	PIMAGE_NT_HEADERS pNtHeader = ::ExGetNtHeader((PIMAGE_DOS_HEADER)dwDosHeader);
 
 	::ExGetSectionUnusedArea(pNtHeader, ".idata", &dwNamePoolAddress, &dwNamePoolSize);
-	//::ExGetNamePool(pNtHeader, &dwNamePoolAddress, &dwNamePoolSize);
 
 	char szSelfFileName[MAX_PATH] = "";
 	::ExGetSelfFileName(szSelfFileName, sizeof(szSelfFileName));
@@ -735,7 +669,7 @@ ExReplaceImportTable(DWORD dwOffset)
 		::ExReplaceImportDllName(dwDosHeader, dwImportTableDelta, &dwNamePoolAddress, &dwNamePoolSize, pDescriptor, "phi.so", g_szPhiSoFileName);
 	}
 
-	DWORD dwOffsetOfName = dwDosHeader - dwImportTableDelta;
+	const DWORD dwOffsetOfName = dwDosHeader - dwImportTableDelta;
 	::ExReplaceImportFunctionName(dwOffsetOfName, pDescriptor, szSelfFileName, "rb_require",   "ex_require");
 	::ExReplaceImportFunctionName(dwOffsetOfName, pDescriptor, szSelfFileName, "rb_f_require", "ex_f_require");
 }
@@ -743,7 +677,7 @@ ExReplaceImportTable(DWORD dwOffset)
 static void
 ExDllWriteToFile(DWORD dwOffset, DWORD dwSize, char *pszFilePath)
 {
-	void  *pvBuffer = (void*)dwOffset;
+	const void *pvBuffer = (void*)dwOffset;
 	DWORD dwWritten = 0;
 
 	HANDLE hFile = ::CreateFile(pszFilePath, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
@@ -756,12 +690,12 @@ ExLoadDll(char *pszFilePath, VALUE vFileName)
 {
 	DEBUGMSG2("ExLoadDll('%s', '%s')\n", pszFilePath, STR2CSTR(vFileName));
 
-	HMODULE hModule = ::LoadLibraryEx(pszFilePath, NULL, LOAD_WITH_ALTERED_SEARCH_PATH);
-	DWORD   dwError = ::GetLastError();
+	const HMODULE hModule = ::LoadLibraryEx(pszFilePath, NULL, LOAD_WITH_ALTERED_SEARCH_PATH);
+	const DWORD   dwError = ::GetLastError();
 
 	if ( hModule ) {
-		VALUE vInitFunctionName = ::ExGetInitFunctionName(vFileName);
-		DWORD dwInitProc = (DWORD)::GetProcAddress(hModule, STR2CSTR(vInitFunctionName));
+		const VALUE vInitFunctionName = ::ExGetInitFunctionName(vFileName);
+		const DWORD dwInitProc = (DWORD)::GetProcAddress(hModule, STR2CSTR(vInitFunctionName));
 
 		if ( dwInitProc ) {
 			DEBUGMSG0("  module info\n");
@@ -790,8 +724,8 @@ ExLoadDll(char *pszFilePath, VALUE vFileName)
 static VALUE
 ExGetInitFunctionName(VALUE vFileName)
 {
-	static VALUE id_basename = ::rb_intern("basename");
-	static VALUE id_sub      = ::rb_intern("sub");
+	static const ID id_basename = ::rb_intern("basename");
+	static const ID id_sub      = ::rb_intern("sub");
 
 	VALUE vBaseName     = ::rb_funcall(rb_cFile, id_basename, 1, vFileName);
 	VALUE vRegexpString = ::rb_str_new2("(\\.so|\\.dll)$");
@@ -805,8 +739,8 @@ ExGetInitFunctionName(VALUE vFileName)
 static VALUE
 ExRequireDll(VALUE vFileName, DWORD dwID)
 {
-	DWORD dwOffset = g_DllOffsetTable[dwID];
-	DWORD dwSize   = g_DllSizeTable[dwID];
+	const DWORD dwOffset = g_DllOffsetTable[dwID];
+	const DWORD dwSize   = g_DllSizeTable[dwID];
 
 	if ( dwOffset && dwSize ) {
 		::ExReplaceImportTable(dwOffset);
@@ -816,7 +750,7 @@ ExRequireDll(VALUE vFileName, DWORD dwID)
 
 		::ExDllWriteToFile(dwOffset, dwSize, szTemporaryFilePath);
 
-		HMODULE hModule = ::ExLoadDll(szTemporaryFilePath, vFileName);
+		const HMODULE hModule = ::ExLoadDll(szTemporaryFilePath, vFileName);
 		if ( hModule ) {
 			DLL_ENTRY* entry = new DLL_ENTRY;
 			entry->hModule     = hModule;
@@ -844,8 +778,8 @@ ExRequireDll(VALUE vFileName, DWORD dwID)
 static VALUE
 ExRequireScript(VALUE vFileName, DWORD dwID)
 {
-	DWORD dwOffset = g_ScriptOffsetTable[dwID];
-	DWORD dwSize   = g_ScriptSizeTable[dwID];
+	const DWORD dwOffset = g_ScriptOffsetTable[dwID];
+	const DWORD dwSize   = g_ScriptSizeTable[dwID];
 
 	if ( dwOffset && dwSize ) {
 		VALUE vScript = ::rb_str_new((char*)dwOffset, dwSize);
@@ -878,37 +812,80 @@ ExGetSelfFileName(LPSTR pszBuffer, UINT uiBufferSize)
 static VALUE
 ExEvalString(VALUE vString, VALUE vFileName)
 {
-	static VALUE id_eval  = ::rb_intern("eval");
-	static VALUE vBinding = ::rb_const_get(rb_mKernel, ::rb_intern("TOPLEVEL_BINDING"));
-	static VALUE vLine    = INT2FIX(1);
+	static const ID    id_eval  = ::rb_intern("eval");
+	static const VALUE vBinding = ::rb_const_get(rb_mKernel, ::rb_intern("TOPLEVEL_BINDING"));
+	static const VALUE vLine    = INT2FIX(1);
 	return ::rb_funcall(rb_mKernel, id_eval, 4, vString, vBinding, vFileName, vLine);
 }
 
+static bool
+ExFindFile(VALUE vFilename, bool *pbShared, VALUE *pvRealname, DWORD *pdwID)
+{
+	const string filename     = STR2CSTR(vFilename);
+	const string filename_rb  = string(filename).append(".rb");
+	const string filename_so  = string(filename).append(".so");
+	const string filename_dll = string(filename).append(".dll");
+
+	*pvRealname = Qnil;
+
+	if ( (*pdwID = g_Name2IdTable[filename]) ) {
+		if ( g_ScriptOffsetTable[*pdwID] ) {
+			*pbShared   = false;
+			*pvRealname = ::rb_str_new2(filename.c_str());
+		} else if ( g_DllOffsetTable[*pdwID] ) {
+			*pbShared   = true;
+			*pvRealname = ::rb_str_new2(filename.c_str());
+		}
+	} else if ( (*pdwID = g_Name2IdTable[filename_rb]) ) {
+		*pbShared   = false;
+		*pvRealname = ::rb_str_new2(filename_rb.c_str());
+	} else if ( (*pdwID = g_Name2IdTable[filename_so]) ) {
+		*pbShared   = true;
+		*pvRealname = ::rb_str_new2(filename_so.c_str());
+	} else if ( (*pdwID = g_Name2IdTable[filename_dll]) ) {
+		*pbShared   = true;
+		*pvRealname = ::rb_str_new2(filename_dll.c_str());
+	}
+
+	return (*pvRealname != Qnil);
+}
+
 static VALUE
 ExRequire(VALUE vObject, VALUE vFileName)
 {
-	VALUE vFeatures = ::rb_gv_get("$\"");
-	if ( ::rb_ary_includes(vFeatures, vFileName) == Qtrue ) {
-		return Qfalse;
-	}
+	bool  bShared   = false;
+	VALUE vRealname = Qnil;
+	DWORD dwID      = 0;
 
-	char *pszFileName  = STR2CSTR(vFileName);
-	char *pszExtension = ::strrchr(pszFileName, '.');
+	if ( ::ExFindFile(vFileName, &bShared, &vRealname, &dwID) ) {
+		DEBUGMSG1("'%s' was found in the archive.\n", STR2CSTR(vFileName));
+		DEBUGMSG2("the realname of '%s' is '%s'.\n", STR2CSTR(vFileName), STR2CSTR(vRealname));
+
+		const VALUE vFeatures = ::rb_gv_get("$\"");
+		if ( ::rb_ary_includes(vFeatures, vRealname) == Qfalse ) {
+			DEBUGMSG1("'%s' was not loaded.\n", STR2CSTR(vFileName));
+
+			::rb_ary_push(vFeatures, vRealname);
+
+			if ( bShared ) {
+				DEBUGMSG1("'%s' is a shared library.\n", STR2CSTR(vFileName));
+				::ExRequireDll(vRealname, dwID);
+			} else {
+				DEBUGMSG1("'%s' is a script.\n", STR2CSTR(vFileName));
+				::ExRequireScript(vRealname, dwID);
+			}
 
-	if ( pszExtension ) {
-		if ( ::ExRequireWithExtension(vFileName, pszExtension) == Qtrue ) {
-			::rb_ary_push(vFeatures, vFileName);
 			return Qtrue;
+		} else {
+			DEBUGMSG1("'%s' was loaded.\n", STR2CSTR(vFileName));
+			return Qfalse;
 		}
 	} else {
-		if ( ::ExRequireWithoutExtension(vFileName) == Qtrue ) {
-			::rb_ary_push(vFeatures, vFileName);
-			return Qtrue;
-		}
+		DEBUGMSG1("'%s' was not found in the archive.\n", STR2CSTR(vFileName));
 	}
 
-	static VALUE id_require = ::rb_intern("__old_require__");
-	return ::rb_funcall(rb_mKernel, id_require, 1, vFileName);
+	static ID id_call = ::rb_intern("call");
+	return ::rb_funcall(g_vOldRequire, id_call, 1, vFileName);
 }
 
 ////////////////////////////////////////////////////////////////////////////////
